#include <sys/queue.h>

#include "spho/spho.h"
#include "spho/tp.h"

#include "msph/sphophi.h"

typedef int (*msph_tree_sphophi_f)(struct spho_scope *, struct msph_tree *);

static int	msph_tree_sphophi_decor_scope(struct spho_scope *,
		    struct msph_tree *);
static int	msph_tree_sphophi_decor_nom_lookup(struct spho_scope *,
		    struct msph_tree *);
static int	msph_tree_sphophi_decor_tp(struct spho_scope *,
		    struct msph_tree *);

/* foreach_pre: top down foreach */
static int	msph_sphophi_foreach_pre(msph_tree_sphophi_f,
		    struct msph_tree_root *);
static int	msph_sphophi_foreach_pre_unit(msph_tree_sphophi_f,
		    struct spho_scope *, struct msph_tree_unit *);
static int	msph_sphophi_foreach_pre_body(msph_tree_sphophi_f, 
		    struct spho_scope *, struct msph_tree_body *);
static int	msph_sphophi_foreach_pre_dir(msph_tree_sphophi_f,
		    struct spho_scope *, struct msph_tree_dir *);
static int	msph_sphophi_foreach_pre_assert(msph_tree_sphophi_f,
		    struct spho_scope *, struct msph_tree_assert *);
static int	msph_sphophi_foreach_pre_membdecl(msph_tree_sphophi_f,
		    struct spho_scope *, struct msph_tree_membdecl *);
static int	msph_sphophi_foreach_pre_tpdef(msph_tree_sphophi_f,
		    struct spho_scope *, struct msph_tree_tpdef *);
static int	msph_sphophi_foreach_pre_nomindecl(msph_tree_sphophi_f,
		    struct spho_scope *, struct msph_tree_nomindecl *);
static int	msph_sphophi_foreach_pre_tpexpr(msph_tree_sphophi_f,
		    struct spho_scope *, struct msph_tree_tpexpr *);
static int	msph_sphophi_foreach_pre_true(msph_tree_sphophi_f,
		    struct spho_scope *, struct msph_tree_true *);
static int	msph_sphophi_foreach_pre_false(msph_tree_sphophi_f,
		    struct spho_scope *, struct msph_tree_false *);
static int	msph_sphophi_foreach_pre_name(msph_tree_sphophi_f,
		    struct spho_scope *, struct msph_tree_name *);
static int	msph_sphophi_foreach_pre_appl(msph_tree_sphophi_f,
		    struct spho_scope *, struct msph_tree_appl *);
static int	msph_sphophi_foreach_pre_trait(msph_tree_sphophi_f,
		    struct spho_scope *, struct msph_tree_trait *);
static int	msph_sphophi_foreach_pre_conj(msph_tree_sphophi_f,
		    struct spho_scope *, struct msph_tree_conj *);
static int	msph_sphophi_foreach_pre_disj(msph_tree_sphophi_f,
		    struct spho_scope *, struct msph_tree_disj *);
static int	msph_sphophi_foreach_pre_impl(msph_tree_sphophi_f,
		    struct spho_scope *, struct msph_tree_impl *);
static int	msph_sphophi_foreach_pre_arrow(msph_tree_sphophi_f,
		    struct spho_scope *, struct msph_tree_arrow *);
static int	msph_sphophi_foreach_pre_box(msph_tree_sphophi_f,
		    struct spho_scope *, struct msph_tree_box *);
static int	msph_sphophi_foreach_pre_forall(msph_tree_sphophi_f,
		    struct spho_scope *, struct msph_tree_forall *);
static int	msph_sphophi_foreach_pre_paren(msph_tree_sphophi_f,
		    struct spho_scope *, struct msph_tree_paren *);
static int	msph_sphophi_foreach_pre_ident(msph_tree_sphophi_f,
		    struct spho_scope *, struct msph_tree_ident *);

/* foreach_post: bottom up foreach */
static int	msph_sphophi_foreach_post(msph_tree_sphophi_f,
		    struct msph_tree_root *);
static int	msph_sphophi_foreach_post_unit(msph_tree_sphophi_f,
		    struct spho_scope *, struct msph_tree_unit *);
static int	msph_sphophi_foreach_post_body(msph_tree_sphophi_f, 
		    struct spho_scope *, struct msph_tree_body *);
static int	msph_sphophi_foreach_post_dir(msph_tree_sphophi_f,
		    struct spho_scope *, struct msph_tree_dir *);
static int	msph_sphophi_foreach_post_assert(msph_tree_sphophi_f,
		    struct spho_scope *, struct msph_tree_assert *);
static int	msph_sphophi_foreach_post_membdecl(msph_tree_sphophi_f,
		    struct spho_scope *, struct msph_tree_membdecl *);
static int	msph_sphophi_foreach_post_tpdef(msph_tree_sphophi_f,
		    struct spho_scope *, struct msph_tree_tpdef *);
static int	msph_sphophi_foreach_post_nomindecl(msph_tree_sphophi_f,
		    struct spho_scope *, struct msph_tree_nomindecl *);
static int	msph_sphophi_foreach_post_tpexpr(msph_tree_sphophi_f,
		    struct spho_scope *, struct msph_tree_tpexpr *);
static int	msph_sphophi_foreach_post_true(msph_tree_sphophi_f,
		    struct spho_scope *, struct msph_tree_true *);
static int	msph_sphophi_foreach_post_false(msph_tree_sphophi_f,
		    struct spho_scope *, struct msph_tree_false *);
static int	msph_sphophi_foreach_post_name(msph_tree_sphophi_f,
		    struct spho_scope *, struct msph_tree_name *);
static int	msph_sphophi_foreach_post_appl(msph_tree_sphophi_f,
		    struct spho_scope *, struct msph_tree_appl *);
static int	msph_sphophi_foreach_post_trait(msph_tree_sphophi_f,
		    struct spho_scope *, struct msph_tree_trait *);
static int	msph_sphophi_foreach_post_conj(msph_tree_sphophi_f,
		    struct spho_scope *, struct msph_tree_conj *);
static int	msph_sphophi_foreach_post_disj(msph_tree_sphophi_f,
		    struct spho_scope *, struct msph_tree_disj *);
static int	msph_sphophi_foreach_post_impl(msph_tree_sphophi_f,
		    struct spho_scope *, struct msph_tree_impl *);
static int	msph_sphophi_foreach_post_arrow(msph_tree_sphophi_f,
		    struct spho_scope *, struct msph_tree_arrow *);
static int	msph_sphophi_foreach_post_box(msph_tree_sphophi_f,
		    struct spho_scope *, struct msph_tree_box *);
static int	msph_sphophi_foreach_post_forall(msph_tree_sphophi_f,
		    struct spho_scope *, struct msph_tree_forall *);
static int	msph_sphophi_foreach_post_paren(msph_tree_sphophi_f,
		    struct spho_scope *, struct msph_tree_paren *);
static int	msph_sphophi_foreach_post_ident(msph_tree_sphophi_f,
		    struct spho_scope *, struct msph_tree_ident *);

int msph_sphophi_init(struct msph_tree_root *root)
{
	if ((root->ctx->spho = spho_ctx_create()) == NULL) {
		MSPH_ERR(root->ctx, MSPH_ERR_SYS);
		return (-1);
	}

	return (0);
}

int
msph_sphophi_scope(struct msph_tree_root *root)
{
	return (msph_sphophi_foreach_pre(msph_tree_sphophi_decor_scope, root));
}

int
msph_sphophi_nom_lookup(struct msph_tree_root *root)
{
	return (msph_sphophi_foreach_pre(msph_tree_sphophi_decor_nom_lookup,
		    root));
}

int
msph_sphophi_tp(struct msph_tree_root *root)
{
	return (msph_sphophi_foreach_post(msph_tree_sphophi_decor_tp, root));
}

int msph_sphophi(struct msph_tree_root *root)
{
	int ret;

	if ((ret = msph_sphophi_init(root)) != 0)
		return (ret);
	if ((ret = msph_sphophi_scope(root)) != 0)
		return (ret);
	if ((ret = msph_sphophi_nom_lookup(root)) != 0)
		return (ret);
	if ((ret = msph_sphophi_tp(root)) != 0)
		return (ret);

	return (ret);
}


#define IS_BINDING(tree)	(((tree)->type == MSPH_TREE_TPDEF)	|| \
				 ((tree)->type == MSPH_TREE_NOMINDECL)	|| \
				 ((tree)->type == MSPH_TREE_MEMBDECL)	|| \
				 ((tree)->type == MSPH_TREE_FORALL))
#define IS_TPEXPR(tree)		(((tree)->type & MSPH_TREE_TPEXPR))

static int
msph_tree_sphophi_decor_scope(struct spho_scope *sc, struct msph_tree *tree)
{
	int ret;
	struct msph_tree_unit *unit;
	struct msph_tree_trait *trait;
	struct msph_tree_forall *forall;
	struct msph_tree_ident *ident;

	ret = 0;

	switch (tree->type) {
	case MSPH_TREE_UNIT:
		unit = (struct msph_tree_unit *)tree;
		if ((unit->dec.sc = spho_scope_create(sc)) == NULL)
			ret = -1;
		break;
	case MSPH_TREE_TRAIT:
		trait = (struct msph_tree_trait *)tree;
		if ((trait->dec.sc = spho_scope_create(sc)) == NULL)
			ret = -1;
		break;
	case MSPH_TREE_FORALL:
		forall = (struct msph_tree_forall *)tree;
		if ((forall->dec.sc = spho_scope_create(sc)) == NULL)
			ret = -1;
		break;
	case MSPH_TREE_IDENT:
		if (! IS_BINDING(tree->parent))
			break;
		ident = (struct msph_tree_ident *)tree;
		if ((ident->dec.nom = spho_scope_nom_add(sc, ident->str,
			    sizeof(ident->str))) == NULL)
			ret = -1;
		break;
	default:
		break;
	}

	return (ret);
}

static int
msph_tree_sphophi_decor_nom_lookup(struct spho_scope *sc,
    struct msph_tree *tree)
{
	int ret;
	struct msph_tree_ident *ident;

	ret = 0;

	switch (tree->type) {
	case MSPH_TREE_IDENT:
		if (tree->parent->type != MSPH_TREE_NAME)
			break;
		ident = (struct msph_tree_ident *)tree;
		if ((ret = spho_scope_nom_lookup(sc, ident->str,
			    sizeof(ident->str), &ident->dec.nom)) != 0)
			break;
		if (ident->dec.nom == NULL)
			ret = -1;
		break;
	default:
		break;
	}

	return (ret);
}

#define TP(ptr)		((ptr)->dec.tp)
#define NOM(ptr)	((ptr)->dec.nom)

static int
msph_tree_sphophi_decor_tp(struct spho_scope *sc, struct msph_tree *tree)
{
	int ret;
	struct msph_tree_tpexpr *tpexpr, *tpexpr_i;
	struct msph_tree_conj *conj;
	struct msph_tree_disj *disj;
	struct msph_tree_impl *impl;
	struct msph_tree_arrow *arrow;
	struct msph_tree_name *name;
	struct msph_tree_appl *appl;
	struct msph_tree_box *box;
	struct msph_tree_forall *forall;
	struct msph_tree_paren *paren;
	struct msph_tree_trait *trait;
	struct msph_tree_dir *dir;
	struct msph_tree_membdecl *membd;

	struct spho_tp_l *tps;
	struct spho_tp *tp, *tp_memb;

	ret = -1;
	tps = NULL;
	tp = tp_memb = NULL;
	dir = NULL;
	membd = NULL;

	if (! (tree->type & MSPH_TREE_TPEXPR))
		return (0);

	tpexpr = (struct msph_tree_tpexpr *)tree;

	switch (tree->type) {
	case MSPH_TREE_CONJ:
		conj = (struct msph_tree_conj *)tree;
		if ((TP(&conj->htpe) = spho_tp_create_conj(sc, TP(conj->ltp),
		    TP(conj->rtp))) == NULL)
			goto cleanup;
		break;
	case MSPH_TREE_DISJ:
		disj = (struct msph_tree_disj *)tree;
		if ((TP(&disj->htpe) = spho_tp_create_conj(sc, TP(disj->ltp),
		    TP(disj->rtp))) == NULL)
			goto cleanup;
		break;
	case MSPH_TREE_IMPL:
		impl = (struct msph_tree_impl *)tree;
		if ((TP(&impl->htpe) = spho_tp_create_conj(sc, TP(impl->ltp),
		    TP(impl->rtp))) == NULL)
			goto cleanup;
		break;
	case MSPH_TREE_ARROW:
		arrow = (struct msph_tree_arrow *)tree;
		if ((TP(&arrow->htpe) = spho_tp_create_conj(sc, TP(arrow->ltp),
		    TP(arrow->rtp))) == NULL)
			goto cleanup;
		break;
	case MSPH_TREE_NAME:
		name = (struct msph_tree_name *)tree;
		if ((TP(&name->htpe) = spho_tp_create_name(sc, NOM(name->id)))
		    == NULL)
			goto cleanup;
		break;
	case MSPH_TREE_APPL:
		appl = (struct msph_tree_appl *)tree;
		if ((tps = malloc(sizeof(*tps))) == NULL) {
			SPHO_SC_ERR(sc, SPHO_ERR_SYS);
			goto cleanup;
		}
		STAILQ_INIT(tps);
		STAILQ_FOREACH(tpexpr_i, &appl->head, entries) {
			STAILQ_INSERT_TAIL(tps, tpexpr_i->dec.tp, entries);
		}
		if ((TP(&appl->htpe) = spho_tp_create_appl(sc, NOM(appl->id),
			    tps)) == NULL)
			goto cleanup;
		tps = NULL;
		break;
	case MSPH_TREE_TRUE:
		if ((TP(tpexpr) = spho_tp_create_true(sc)) == NULL)
			goto cleanup;
		break;
	case MSPH_TREE_FALSE:
		if ((TP(tpexpr) = spho_tp_create_false(sc)) == NULL)
			goto cleanup;
		break;
	case MSPH_TREE_FORALL:
		forall = (struct msph_tree_forall *)tree;
		if ((TP(&forall->htpe) = spho_tp_create_forall(sc,
			    NOM(forall->id), TP(forall->inner))) == NULL)
			goto cleanup;
		break;
	case MSPH_TREE_BOX:
		box = (struct msph_tree_box *)tree;
		if ((TP(&box->htpe) = spho_tp_create_box(sc, TP(box->inner)))
		    == NULL)
			goto cleanup;
		break;
	case MSPH_TREE_PAREN:
		paren = (struct msph_tree_paren *)tree;
		if ((TP(&paren->htpe) = TP(paren->inner)) == NULL)
			goto cleanup;
		break;
	case MSPH_TREE_TRAIT:
		trait = (struct msph_tree_trait *)tree;

		STAILQ_FOREACH(dir, &trait->body->head, entries) {
			if (T(dir)->type != MSPH_TREE_MEMBDECL)
				continue;
			membd = (struct msph_tree_membdecl *)dir;
			// TODO check that we make sure to dealloc when
			// spho has stabilized a bit
			if ((tp_memb = spho_tp_create_member(sc, membd->id->dec.nom,
			    membd->tp->dec.tp)) == NULL)
				goto cleanup;
			if (tp == NULL) {
				tp = tp_memb;
			} else {
				if ((tp = spho_tp_create_conj(sc, tp, tp_memb))
				    == NULL)
					goto cleanup;
			}
			trait->htpe.dec.tp = tp;
		}
		break;
	default:
		SPHO_ASSERT(0);
		break;
	}

	ret = 0;

cleanup:
	if (tps != NULL)
		free(tps);

	return (ret);
}

/* foreach pre: top down foreach node */
static int
msph_sphophi_foreach_pre(msph_tree_sphophi_f f, struct msph_tree_root *root)
{
	int ret;
	struct spho_scope *sc;
	struct msph_tree_unit *unit;

	if (root == NULL || root->ctx == NULL)
		return (-1);

	if (root->ctx->spho == NULL) {
		MSPH_ERR(root->ctx, MSPH_ERR_TYPE_INVAL);
		return (-1);
	}

	sc = spho_scope_global(root->ctx->spho);

	if ((ret = f(sc, T(root))) != 0) {
		goto err;
	}

	ret = 0;
	STAILQ_FOREACH(unit, &root->head, entries) {
		if ((ret = msph_sphophi_foreach_pre_unit(f, sc, unit)) != 0)
			goto err;
	}

	return (ret);
err:
	MSPH_ERR(root->ctx, MSPH_ERR_SPHOPHI);
	return (ret);
}

static int
msph_sphophi_foreach_pre_unit(msph_tree_sphophi_f f, struct spho_scope *sc,
    struct msph_tree_unit *unit)
{
	int ret;

	if ((ret = f(sc, T(unit))) != 0)
		return (ret);

	sc = GET_SCOPE(unit);

	return (msph_sphophi_foreach_pre_body(f, sc, unit->body));
}

static int
msph_sphophi_foreach_pre_body(msph_tree_sphophi_f f, struct spho_scope *sc,
    struct msph_tree_body *body)
{
	int ret;
	struct msph_tree_dir *dir;

	if ((ret = f(sc, T(body))) != 0)
		return (ret);

	ret = 0;
	STAILQ_FOREACH(dir, &body->head, entries) {
		if ((ret = msph_sphophi_foreach_pre_dir(f, sc, dir)) != 0)
			return (ret);
	}

	return (ret);
}

static int
msph_sphophi_foreach_pre_dir(msph_tree_sphophi_f f, struct spho_scope *sc,
    struct msph_tree_dir *dir)
{
	int ret;

	ret = -1;
	switch (dir->htxt.htr.type) {
	case MSPH_TREE_TPDEF:
		ret = msph_sphophi_foreach_pre_tpdef(f, sc,
		    (struct msph_tree_tpdef *)dir);
		break;
	case MSPH_TREE_MEMBDECL:
		ret = msph_sphophi_foreach_pre_membdecl(f, sc,
		    (struct msph_tree_membdecl *)dir);
		break;
	case MSPH_TREE_NOMINDECL:
		ret = msph_sphophi_foreach_pre_nomindecl(f, sc,
		    (struct msph_tree_nomindecl *)dir);
		break;
	case MSPH_TREE_ASSERT:
		ret = msph_sphophi_foreach_pre_assert(f, sc,
		    (struct msph_tree_assert *)dir);
		break;
	default:
		SPHO_ASSERT(0);
		break;
	}

	return (ret);
}

static int
msph_sphophi_foreach_pre_tpdef(msph_tree_sphophi_f f, struct spho_scope *sc,
    struct msph_tree_tpdef *tpdef)
{
	int ret;

	if ((ret = f(sc, T(tpdef))) != 0)
		return (ret);

	if ((ret = msph_sphophi_foreach_pre_ident(f, sc, tpdef->id)) != 0)
		return (ret);

	return (msph_sphophi_foreach_pre_tpexpr(f, sc, tpdef->tp));
}

static int
msph_sphophi_foreach_pre_nomindecl(msph_tree_sphophi_f f, struct spho_scope *sc,
    struct msph_tree_nomindecl *nomd)
{
	int ret;

	if ((ret = f(sc, T(nomd))) != 0)
		return (ret);

	return (msph_sphophi_foreach_pre_ident(f, sc, nomd->id));
}

static int
msph_sphophi_foreach_pre_membdecl(msph_tree_sphophi_f f, struct spho_scope *sc,
    struct msph_tree_membdecl *membd)
{
	int ret;

	if ((ret = f(sc, T(membd))) != 0)
		return (ret);

	if (msph_sphophi_foreach_pre_ident(f, sc, membd->id) != 0)
		return (ret);
	if (msph_sphophi_foreach_pre_tpexpr(f, sc, membd->tp) != 0)
		return (ret);

	return (0);
}

static int
msph_sphophi_foreach_pre_assert(msph_tree_sphophi_f f, struct spho_scope *sc,
    struct msph_tree_assert *ass)
{
	int ret;

	if ((ret = f(sc, T(ass))) != 0)
		return (ret);

	if ((ret = msph_sphophi_foreach_pre_tpexpr(f, sc, ass->ltp)) != 0)
		return (ret);
	if ((ret = msph_sphophi_foreach_pre_tpexpr(f, sc, ass->rtp)) != 0)
		return (ret);

	return (ret);
}

static int
msph_sphophi_foreach_pre_ident(msph_tree_sphophi_f f, struct spho_scope *sc,
    struct msph_tree_ident *ident)
{
	return (f(sc, T(ident)));
}

static int
msph_sphophi_foreach_pre_tpexpr(msph_tree_sphophi_f f, struct spho_scope *sc,
    struct msph_tree_tpexpr *tpexpr)
{
	int ret;

	ret = -1;

	switch (tpexpr->htxt.htr.type) {
	case MSPH_TREE_CONJ:
		ret = msph_sphophi_foreach_pre_conj(f, sc,
		    (struct msph_tree_conj *)tpexpr);
		break;
	case MSPH_TREE_DISJ:
		ret = msph_sphophi_foreach_pre_disj(f, sc,
		    (struct msph_tree_disj *)tpexpr);
		break;
	case MSPH_TREE_IMPL:
		ret = msph_sphophi_foreach_pre_impl(f, sc,
		    (struct msph_tree_impl *)tpexpr);
		break;
	case MSPH_TREE_ARROW:
		ret = msph_sphophi_foreach_pre_arrow(f, sc,
		    (struct msph_tree_arrow *)tpexpr);
		break;
	case MSPH_TREE_NAME:
		ret = msph_sphophi_foreach_pre_name(f, sc,
		    (struct msph_tree_name *)tpexpr);
		break;
	case MSPH_TREE_APPL:
		ret = msph_sphophi_foreach_pre_appl(f, sc,
		    (struct msph_tree_appl *)tpexpr);
		break;
	case MSPH_TREE_TRAIT:
		ret = msph_sphophi_foreach_pre_trait(f, sc,
		    (struct msph_tree_trait *)tpexpr);
		break;
	case MSPH_TREE_BOX:
		ret = msph_sphophi_foreach_pre_box(f, sc,
		    (struct msph_tree_box *)tpexpr);
		break;
	case MSPH_TREE_FORALL:
		ret = msph_sphophi_foreach_pre_forall(f, sc,
		    (struct msph_tree_forall *)tpexpr);
		break;
	case MSPH_TREE_PAREN:
		ret = msph_sphophi_foreach_pre_paren(f, sc,
		    (struct msph_tree_paren *)tpexpr);
		break;
	case MSPH_TREE_TRUE:
		ret = msph_sphophi_foreach_pre_true(f, sc,
		    (struct msph_tree_true *)tpexpr);
		break;
	case MSPH_TREE_FALSE:
		ret = msph_sphophi_foreach_pre_false(f, sc,
		    (struct msph_tree_false *)tpexpr);
		break;
	default:
		SPHO_ASSERT(0);
		break;
	}

	return (ret);
}

static int
msph_sphophi_foreach_pre_conj(msph_tree_sphophi_f f, struct spho_scope *sc,
    struct msph_tree_conj *conj)
{
	int ret;

	if ((ret = f(sc, T(conj))) != 0)
		return (ret);

	if ((ret = msph_sphophi_foreach_pre_tpexpr(f, sc, conj->ltp)) != 0)
		return (ret);

	return (msph_sphophi_foreach_pre_tpexpr(f, sc, conj->rtp));
}

static int
msph_sphophi_foreach_pre_disj(msph_tree_sphophi_f f, struct spho_scope *sc,
    struct msph_tree_disj *disj)
{
	int ret;

	if ((ret = f(sc, T(disj))) != 0)
		return (ret);
	if ((ret = msph_sphophi_foreach_pre_tpexpr(f, sc, disj->ltp)) != 0)
		return (ret);
	return (msph_sphophi_foreach_pre_tpexpr(f, sc, disj->rtp));
}

static int
msph_sphophi_foreach_pre_impl(msph_tree_sphophi_f f, struct spho_scope *sc,
    struct msph_tree_impl *impl)
{
	int ret;

	if ((ret = f(sc, T(impl))) != 0)
		return (ret);

	if ((ret = msph_sphophi_foreach_pre_tpexpr(f, sc, impl->ltp)) != 0)
		return (ret);
	return (msph_sphophi_foreach_pre_tpexpr(f, sc, impl->rtp));
}

static int
msph_sphophi_foreach_pre_arrow(msph_tree_sphophi_f f, struct spho_scope *sc,
    struct msph_tree_arrow *arrow)
{
	int ret;

	if ((ret = f(sc, T(arrow))) != 0)
		return (ret);

	if ((ret = msph_sphophi_foreach_pre_tpexpr(f, sc, arrow->ltp)) != 0)
		return (ret);
	return (msph_sphophi_foreach_pre_tpexpr(f, sc, arrow->rtp));
}

static int
msph_sphophi_foreach_pre_name(msph_tree_sphophi_f f, struct spho_scope *sc,
    struct msph_tree_name *name)
{
	int ret;

	if ((ret = f(sc, T(name))) != 0)
		return (ret);

	return (msph_sphophi_foreach_pre_ident(f, sc, name->id));
}

static int
msph_sphophi_foreach_pre_appl(msph_tree_sphophi_f f, struct spho_scope *sc,
    struct msph_tree_appl *appl)
{
	int ret;
	struct msph_tree_tpexpr *tp;

	if ((ret = f(sc, T(appl))) != 0)
		return (ret);

	ret = 0;
	STAILQ_FOREACH(tp, &appl->head, entries) {
		if ((ret = msph_sphophi_foreach_pre_tpexpr(f, sc, tp)) != 0)
			return (ret);
	}

	return (ret);
}

static int
msph_sphophi_foreach_pre_trait(msph_tree_sphophi_f f, struct spho_scope *sc,
    struct msph_tree_trait *trait)
{
	int ret;
	if ((ret = f(sc, T(trait))) != 0)
		return (ret);

	sc = GET_SCOPE(trait);

	return (msph_sphophi_foreach_pre_body(f, sc, trait->body));
}

static int
msph_sphophi_foreach_pre_box(msph_tree_sphophi_f f,struct spho_scope *sc,
    struct msph_tree_box *box)
{
	int ret;

	if ((ret = f(sc, T(box))) != 0)
		return (ret);

	return (msph_sphophi_foreach_pre_tpexpr(f, sc, box->inner));
}

static int
msph_sphophi_foreach_pre_forall(msph_tree_sphophi_f f, struct spho_scope *sc,
    struct msph_tree_forall *forall)
{
	int ret;
	
	if ((ret = f(sc, T(forall))) != 0)
		return (ret);

	sc = GET_SCOPE(forall);

	return (msph_sphophi_foreach_pre_tpexpr(f, sc, forall->inner));
}

static int
msph_sphophi_foreach_pre_paren(msph_tree_sphophi_f f, struct spho_scope *sc,
    struct msph_tree_paren *paren)
{
	int ret;

	if ((ret = f(sc, T(paren))) != 0)
		return (ret);

	return (msph_sphophi_foreach_pre_tpexpr(f, sc, paren->inner));
}

static int
msph_sphophi_foreach_pre_true(msph_tree_sphophi_f f, struct spho_scope *sc,
    struct msph_tree_true *tru)
{
	return (f(sc, T(tru)));
}

static int
msph_sphophi_foreach_pre_false(msph_tree_sphophi_f f, struct spho_scope *sc,
    struct msph_tree_false *fls)
{
	return (f(sc, T(fls)));
}


/* foreach post: bottom up foreach */
static int
msph_sphophi_foreach_post(msph_tree_sphophi_f f, struct msph_tree_root *root)
{
	int ret;
	struct spho_scope *sc;
	struct msph_tree_unit *unit;

	if (root == NULL || root->ctx == NULL)
		return (-1);

	if (root->ctx->spho == NULL) {
		MSPH_ERR(root->ctx, MSPH_ERR_TYPE_INVAL);
		return (-1);
	}

	sc = spho_scope_global(root->ctx->spho);

	ret = 0;
	STAILQ_FOREACH(unit, &root->head, entries) {
		if ((ret = msph_sphophi_foreach_post_unit(f, sc, unit)) != 0)
			goto err;
	}

	if ((ret = f(sc, T(root))) != 0)
		goto err;

	return (ret);
err:
	MSPH_ERR(root->ctx, MSPH_ERR_SPHOPHI);
	return (ret);
}

static int
msph_sphophi_foreach_post_unit(msph_tree_sphophi_f f, struct spho_scope *sc,
    struct msph_tree_unit *unit)
{
	int ret;
	struct spho_scope *inner_sc;

	inner_sc = GET_SCOPE(unit);

	if ((ret = msph_sphophi_foreach_post_body(f, inner_sc, unit->body))
	    != 0)
		return (ret);

	return (f(sc, T(unit)));
}

static int
msph_sphophi_foreach_post_body(msph_tree_sphophi_f f, struct spho_scope *sc,
    struct msph_tree_body *body)
{
	int ret;
	struct msph_tree_dir *dir;

	ret = 0;
	STAILQ_FOREACH(dir, &body->head, entries) {
		if ((ret = msph_sphophi_foreach_post_dir(f, sc, dir)) != 0)
			return (ret);
	}

	return (f(sc, T(body)));
}

static int
msph_sphophi_foreach_post_dir(msph_tree_sphophi_f f, struct spho_scope *sc,
    struct msph_tree_dir *dir)
{
	int ret;

	ret = -1;
	switch (dir->htxt.htr.type) {
	case MSPH_TREE_TPDEF:
		ret = msph_sphophi_foreach_post_tpdef(f, sc,
		    (struct msph_tree_tpdef *)dir);
		break;
	case MSPH_TREE_MEMBDECL:
		ret = msph_sphophi_foreach_post_membdecl(f, sc,
		    (struct msph_tree_membdecl *)dir);
		break;
	case MSPH_TREE_NOMINDECL:
		ret = msph_sphophi_foreach_post_nomindecl(f, sc,
		    (struct msph_tree_nomindecl *)dir);
		break;
	case MSPH_TREE_ASSERT:
		ret = msph_sphophi_foreach_post_assert(f, sc,
		    (struct msph_tree_assert *)dir);
		break;
	default:
		SPHO_ASSERT(0);
		break;
	}

	return (ret);
}

static int
msph_sphophi_foreach_post_tpdef(msph_tree_sphophi_f f, struct spho_scope *sc,
    struct msph_tree_tpdef *tpdef)
{
	int ret;

	if ((ret = msph_sphophi_foreach_post_ident(f, sc, tpdef->id)) != 0)
		return (ret);
	if ((ret = msph_sphophi_foreach_post_tpexpr(f, sc, tpdef->tp)) != 0)
		return (ret);

	return (f(sc, T(tpdef)));
}

static int
msph_sphophi_foreach_post_nomindecl(msph_tree_sphophi_f f,
    struct spho_scope *sc, struct msph_tree_nomindecl *nomd)
{
	int ret;

	if ((ret = msph_sphophi_foreach_post_ident(f, sc, nomd->id)) != 0)
		return (ret);

	return (f(sc, T(nomd)));
}

static int
msph_sphophi_foreach_post_membdecl(msph_tree_sphophi_f f, struct spho_scope *sc,
    struct msph_tree_membdecl *membd)
{
	int ret;

	if ((ret = msph_sphophi_foreach_post_ident(f, sc, membd->id)) != 0)
		return (ret);
	if ((ret = msph_sphophi_foreach_post_tpexpr(f, sc, membd->tp)) != 0)
		return (ret);

	return (f(sc, T(membd)));
}

static int
msph_sphophi_foreach_post_assert(msph_tree_sphophi_f f, struct spho_scope *sc,
    struct msph_tree_assert *ass)
{
	int ret;

	if ((ret = msph_sphophi_foreach_post_tpexpr(f, sc, ass->ltp)) != 0)
		return (ret);
	if ((ret = msph_sphophi_foreach_post_tpexpr(f, sc, ass->rtp)) != 0)
		return (ret);

	return (f(sc, T(ass)));
}

static int
msph_sphophi_foreach_post_ident(msph_tree_sphophi_f f, struct spho_scope *sc,
    struct msph_tree_ident *ident)
{
	return (f(sc, T(ident)));
}

static int
msph_sphophi_foreach_post_tpexpr(msph_tree_sphophi_f f, struct spho_scope *sc,
    struct msph_tree_tpexpr *tpexpr)
{
	int ret;

	ret = -1;

	switch (tpexpr->htxt.htr.type) {
	case MSPH_TREE_CONJ:
		ret = msph_sphophi_foreach_post_conj(f, sc,
		    (struct msph_tree_conj *)tpexpr);
		break;
	case MSPH_TREE_DISJ:
		ret = msph_sphophi_foreach_post_disj(f, sc,
		    (struct msph_tree_disj *)tpexpr);
		break;
	case MSPH_TREE_IMPL:
		ret = msph_sphophi_foreach_post_impl(f, sc,
		    (struct msph_tree_impl *)tpexpr);
		break;
	case MSPH_TREE_ARROW:
		ret = msph_sphophi_foreach_post_arrow(f, sc,
		    (struct msph_tree_arrow *)tpexpr);
		break;
	case MSPH_TREE_NAME:
		ret = msph_sphophi_foreach_post_name(f, sc,
		    (struct msph_tree_name *)tpexpr);
		break;
	case MSPH_TREE_APPL:
		ret = msph_sphophi_foreach_post_appl(f, sc,
		    (struct msph_tree_appl *)tpexpr);
		break;
	case MSPH_TREE_TRAIT:
		ret = msph_sphophi_foreach_post_trait(f, sc,
		    (struct msph_tree_trait *)tpexpr);
		break;
	case MSPH_TREE_BOX:
		ret = msph_sphophi_foreach_post_box(f, sc,
		    (struct msph_tree_box *)tpexpr);
		break;
	case MSPH_TREE_FORALL:
		ret = msph_sphophi_foreach_post_forall(f, sc,
		    (struct msph_tree_forall *)tpexpr);
		break;
	case MSPH_TREE_PAREN:
		ret = msph_sphophi_foreach_post_paren(f, sc,
		    (struct msph_tree_paren *)tpexpr);
		break;
	case MSPH_TREE_TRUE:
		ret = msph_sphophi_foreach_post_true(f, sc,
		    (struct msph_tree_true *)tpexpr);
		break;
	case MSPH_TREE_FALSE:
		ret = msph_sphophi_foreach_post_false(f, sc,
		    (struct msph_tree_false *)tpexpr);
		break;
	default:
		SPHO_ASSERT(0);
		break;
	}

	return (ret);
}

static int
msph_sphophi_foreach_post_conj(msph_tree_sphophi_f f, struct spho_scope *sc,
    struct msph_tree_conj *conj)
{
	int ret;

	if ((ret = msph_sphophi_foreach_post_tpexpr(f, sc, conj->ltp)) != 0)
		return (ret);
	if ((ret = msph_sphophi_foreach_post_tpexpr(f, sc, conj->rtp)) != 0)
		return (ret);

	return (f(sc, T(conj)));

}

static int
msph_sphophi_foreach_post_disj(msph_tree_sphophi_f f, struct spho_scope *sc,
    struct msph_tree_disj *disj)
{
	int ret;

	if ((ret = msph_sphophi_foreach_post_tpexpr(f, sc, disj->ltp)) != 0)
		return (ret);
	if ((ret = msph_sphophi_foreach_post_tpexpr(f, sc, disj->rtp)) != 0)
		return (ret);

	return (f(sc, T(disj)));
}

static int
msph_sphophi_foreach_post_impl(msph_tree_sphophi_f f, struct spho_scope *sc,
    struct msph_tree_impl *impl)
{
	int ret;

	if ((ret = msph_sphophi_foreach_post_tpexpr(f, sc, impl->ltp)) != 0)
		return (ret);
	if ((ret = msph_sphophi_foreach_post_tpexpr(f, sc, impl->rtp)) != 0)
		return (ret);

	return (f(sc, T(impl)));
}

static int
msph_sphophi_foreach_post_arrow(msph_tree_sphophi_f f, struct spho_scope *sc,
    struct msph_tree_arrow *arrow)
{
	int ret;

	if ((ret = msph_sphophi_foreach_post_tpexpr(f, sc, arrow->ltp)) != 0)
		return (ret);
	if ((ret = msph_sphophi_foreach_post_tpexpr(f, sc, arrow->rtp)) != 0)
		return (ret);

	return (f(sc, T(arrow)));
}

static int
msph_sphophi_foreach_post_name(msph_tree_sphophi_f f, struct spho_scope *sc,
    struct msph_tree_name *name)
{
	int ret;

	if ((ret = msph_sphophi_foreach_post_ident(f, sc, name->id)) != 0)
		return (ret);

	return (f(sc, T(name)));

}

static int
msph_sphophi_foreach_post_appl(msph_tree_sphophi_f f, struct spho_scope *sc,
    struct msph_tree_appl *appl)
{
	int ret;
	struct msph_tree_tpexpr *tp;

	ret = 0;
	STAILQ_FOREACH(tp, &appl->head, entries) {
		if ((ret = msph_sphophi_foreach_post_tpexpr(f, sc, tp)) != 0)
			return (ret);
	}

	return (f(sc, T(appl)));
}

static int
msph_sphophi_foreach_post_trait(msph_tree_sphophi_f f, struct spho_scope *sc,
    struct msph_tree_trait *trait)
{
	int ret;
	struct spho_scope *inner_sc;

	inner_sc = GET_SCOPE(trait);

	if ((ret = msph_sphophi_foreach_post_body(f, inner_sc, trait->body))
	    != 0)
		return (ret);

	return (f(sc, T(trait)));
}

static int
msph_sphophi_foreach_post_box(msph_tree_sphophi_f f,struct spho_scope *sc,
    struct msph_tree_box *box)
{
	int ret;

	if ((ret = msph_sphophi_foreach_post_tpexpr(f, sc, box->inner)) != 0)
		return (ret);

	return (f(sc, T(box)));
}

static int
msph_sphophi_foreach_post_forall(msph_tree_sphophi_f f, struct spho_scope *sc,
    struct msph_tree_forall *forall)
{
	int ret;
	struct spho_scope *inner_sc;
	
	inner_sc = GET_SCOPE(forall);

	if ((ret = msph_sphophi_foreach_post_tpexpr(f, inner_sc, forall->inner))
	    != 0)
		return (ret);

	return (f(sc, T(forall)));

}

static int
msph_sphophi_foreach_post_paren(msph_tree_sphophi_f f, struct spho_scope *sc,
    struct msph_tree_paren *paren)
{
	int ret;

	if ((ret = msph_sphophi_foreach_post_tpexpr(f, sc, paren->inner)) != 0)
		return (ret);

	return (f(sc, T(paren)));

}

static int
msph_sphophi_foreach_post_true(msph_tree_sphophi_f f, struct spho_scope *sc,
    struct msph_tree_true *tru)
{
	return (f(sc, T(tru)));
}

static int
msph_sphophi_foreach_post_false(msph_tree_sphophi_f f, struct spho_scope *sc,
    struct msph_tree_false *fls)
{
	return (f(sc, T(fls)));
}
